import setuptools

setuptools.setup()

# [quote](https://setuptools.readthedocs.io/en/latest/setuptools.html)
# "As PEP 517 is new, support is not universal, and frontends that do
# support it may still have bugs. For compatibility, you may want to
# put a setup.py file containing only a setuptools.setup() invocation."

